/************************************************************************/
/*                                                                      */
/*   knn.h                                                              */
/*                                                                      */
/*   Definitions and functions used by both svm_learn and svm_classify. */
/*                                                                      */
/*   Copyright: Thorsten Joachims                                       */
/*   Date: 14.10.99                                                     */
/*                                                                      */
/************************************************************************/

# include <stdio.h>
# include <ctype.h>
# include <signal.h>
# include <math.h>
# include <string.h>
# include <stdlib.h>

char version[] = "V1.0";
char version_date[] = "22.12.97";

typedef struct word {
  long    wnum;	
  double  weight;
} WORD;

typedef struct kernel_parm {
  long    kernel_type;   /* 0=linear, 1=poly, 2=rbf, 3=sigmoid, 4=custom */
  long    poly_degree;
  double  rbf_gamma;
  double  sigmoid_gamma;
  double  sigmoid_c;
  char    custom[50];    /* for user supplied kernel */
} KERNEL_PARM;

typedef struct model {
  long    sv_num;	
  double  b;
  WORD  **supvec;
  double  *alpha;
  long    *docnum;
} MODEL;

double classify_example();
double kernel(); 
double custom_kernel(); 
double sprod_ss();
long min();
long max();
double model_length_s();

long verbosity=2;              /* verbosity level (0-3) */

# include "kernel.h"         /* this contains a user supplied kernel */

double classify_example(model,ex,kernel_parm) /* classifies example */
MODEL *model;
WORD *ex;
KERNEL_PARM *kernel_parm;
{
  register long i;
  register double dist=0;

  for(i=1;i<model->sv_num;i++) {  
    dist+=kernel(kernel_parm,model->supvec[i],ex)*model->alpha[i];
  }
  return(dist-model->b);
}

double kernel(kernel_parm,a,b) /* calculate the kernel function */
KERNEL_PARM *kernel_parm;
WORD *a,*b;
{
  switch(kernel_parm->kernel_type) {
    case 0: /* linear */ 
            return(sprod_ss(a,b)); 
    case 1: /* polynomial */
            return(pow(sprod_ss(a,b)+1,(double)kernel_parm->poly_degree)); 
    case 2: /* radial basis function */
            return(exp(-kernel_parm->rbf_gamma*(sprod_ss(a,a)-2*sprod_ss(a,b)+sprod_ss(b,b)))); 
    case 3: /* sigmoid neural net */
            return(tanh(kernel_parm->sigmoid_gamma*sprod_ss(a,b)+kernel_parm->sigmoid_c)); 
    case 4: /* custom-kernel supplied in file kernel.h*/
            return(custom_kernel(kernel_parm,a,b)); 
    default: printf("Error: Unknown kernel function\n");
  }
}

double sprod_ss(a,b) /* compute the inner product of two sparse vectors */
WORD *a,*b;          
{
    register long i=0,j=0,step;
    register double sum=0;
    register WORD *ai,*bj;
    step=sizeof(WORD);
    ai=a;
    bj=b;
    while (ai->wnum && bj->wnum) {
      if(ai->wnum > bj->wnum) {
	bj=(WORD *)((long)bj+step);
      }
      else if (ai->wnum < bj->wnum) {
	ai=(WORD *)((long)ai+step);
      }
      else {
	sum+=ai->weight * bj->weight;
	ai=(WORD *)((long)ai+step);
	bj=(WORD *)((long)bj+step);
      }
    }
    return(sum);
}

double model_length_s(model,kernel_parm) /* compute length of weight vector */
MODEL *model;
KERNEL_PARM *kernel_parm;
{
  register long i,j;
  register double sum=0,alphai;
  register WORD *supveci;

  for(i=1;i<model->sv_num;i++) {  
    alphai=model->alpha[i];
    supveci=model->supvec[i];
    for(j=1;j<model->sv_num;j++) {
      sum+=alphai*model->alpha[j]
	   *kernel(kernel_parm,supveci,model->supvec[j]);
    }
  }
  return(sqrt(sum));
}

read_documents(docfile,docs,label,max_words_doc,ll,totwords,totdoc)
char *docfile;
WORD **docs;
long *label;
long max_words_doc,ll,*totwords,*totdoc;
{
  char *line;
  WORD *doc;
  long dnum=0,wpos,i,j;
  long doc_label;
  FILE *docfl;

  line = (char *)malloc(sizeof(char)*ll);

  if ((docfl = fopen (docfile, "r")) == NULL)
  { perror (docfile); exit (1); }

  fscanf(docfl,"%[^\n]\n", line);  /* first line contains comments */

  doc = (WORD *)malloc(sizeof(WORD)*(max_words_doc+10));
  if(verbosity>=1) {
    printf("Reading documents into memory..."); fflush(stdout);
  }
  dnum=0;
  (*totwords)=0;
  while(fscanf(docfl,"%[^\n]\n", line) != EOF) {
    parse_document(line,doc,&doc_label,&wpos,max_words_doc);
    label[dnum]=doc_label;
    if(doc[wpos-2].wnum>(*totwords)) (*totwords)=doc[wpos-2].wnum;
    if(!(docs[dnum] = (WORD *)malloc(sizeof(WORD)*wpos)))
      { perror ("Could not malloc!!!\n"); exit (1); }
    for(i=0;i<wpos;i++) (docs[dnum])[i]=doc[i];
    dnum++;
    if(verbosity>=1) {
      if((dnum % 100) == 0) {
	printf("%ld..",dnum); fflush(stdout);
      }
    }
  }
  fclose(docfl);
  free(line);
  free(doc);
  if(verbosity>=1) {
    fprintf(stdout, "OK. (%d documents read)\n", dnum);
  }
  (*totdoc)=dnum;
}

parse_document(line,doc,label,numwords,max_words_doc)
char *line;
WORD *doc;
long *label,*numwords,max_words_doc;
{
  register long wpos,i,j,pos;
  long wnum;
  double weight;

  pos=0;
  wpos=0;
  sscanf(line,"%ld",label);
  while(line[++pos]>32);
  while((sscanf(line+pos,"%ld:%lf",&wnum,&weight) != EOF) && 
	(wpos<max_words_doc)) {
    while(line[++pos]>32);
    if(wnum<=0) { 
      perror ("Feature numbers must be larger or equal to 1!!!\n"); 
      exit (1); 
    }
    doc[wpos].wnum=wnum;
    doc[wpos].weight=weight; 
    wpos++;
  }
  doc[wpos].wnum=0;
  (*numwords)=wpos+1;
}


long min(a,b)
long a,b;
{
  if(a<b)
    return(a);
  else
    return(b);
}

long max(a,b)
long a,b;
{
  if(a>b)
    return(a);
  else
    return(b);
}

nol_ll(file,nol,wol,ll) /* grep through file and count number of lines, */
char *file;             /* maximum number of spaces per line, and longest */
long *nol,*wol,*ll;     /* line. */
{
  FILE *fl;
  char c;
  long current_length,current_wol;

  if ((fl = fopen (file, "r")) == NULL)
  { perror (file); exit (1); }
  current_length=0;
  current_wol=0;
  (*ll)=0;
  (*nol)=1;
  (*wol)=0;
  while((c=getc(fl)) != EOF) {
    current_length++;
    if(c == ' ') {
      current_wol++;
    }
    if(c == '\n') {
      (*nol)++;
      if(current_length>(*ll)) {
	(*ll)=current_length;
      }
      if(current_wol>(*wol)) {
	(*wol)=current_wol;
      }
      current_length=0;
      current_wol=0;
    }
  }
  fclose(fl);
}

copyright_notice()
{
  printf("\nCopyright: Thorsten Joachims, thorsten@ls8.cs.uni-dortmund.de\n\n");
  printf("This software is available for non-commercial use only. It must not\n");
  printf("be modified or distributed without prior permission of the author.\n\n");
}
